/*								*/



/*								*/



/*								*/



/*								*/																		

/*			Permission to use				  */	



/*  Sony Computer Entertainment supplies this lws2rsd software free 	  */

/*  of charge to licensees along with the source code files.		  */



/*  This lws2rsd software is experimental and no support of any kind is 	  */

/*  available.							  */



/*  The licensee is hereby granted permission to use, copy, modify, and	  */

/*  distribute this lws2rsd software and documentation for any purpose. 	  */

/*  Sony Computer Entertainment shall bear no responsibility for any 	  */

/*  damage or loss sustained either directly or indirectly as a result of	  */ 

/*  using this software. 						  */



/*  Copyright 1997 by Sony Computer Entertainment Inc. All Rights Reserved. */





#include <stdlib.h>

#include <string.h>



#include "vecmat.h"

#include "lw.h"



#define NUM_VERTEX 10					/* Number of vertices in one line */

#define COD_ID "@COD970401"						/* COD File ID */

#define RATE 4096.0								/* 4096 = 1 */

#define BUFFER_LENGTH 256					/* Size of character buffer */



struct VERTEX{

	Scalar x;

	Scalar y;

	Scalar z;

	Scalar r;			/* Distance of the center of gravity of each Bone */

	int cno;			/* Coordinate No. */

} *malvert;



struct COORDINATE{

	Param center;

	Scalar m[3][3];

	Scalar t[3];

	Scalar x;

	Scalar y;

	Scalar z;

	int super;

	int numvert;		/* Number of assigned vertices */

} *malcoord;



char Buffer[BUFFER_LENGTH];						/* Length of character buffer */

int NumVert;									/* Number of vertices */

int NumCoord;									/* Number of coordinates */



char *get1line(FILE *fp);				/* Function which reads one valid line */



int createCOD(char *plyName, int NumBone, BonePtr bones, char *codName)

{

	int error_no = 0;	/* Error No. */

	struct VERTEX *vert;

	struct COORDINATE *coord;

	FILE *fp;

	int i, j, k;

	Matrix matrix;

	Object object;

	Param start;

	Param direction;

	Scalar length;

	Param center;

	int counter1, counter2;



	/* Open PLY file */

	fp = fopen(plyName, "r");

	if(!fp){

		error_no = 1;	/* PLYfile fopen error */

		goto createCOD_end;

	}



	/* Read ID */

	get1line(fp);



	/* Check file type */

	if(strncmp(Buffer, "@PLY", 4)) {

		error_no = 2;	/* PLYfile type error */

		goto createCOD_end;

	}



	/* Check buffer */

	if(strncmp(Buffer + 4, "940102", 6)) {

		error_no = 3;	/* PLYfile version error */

		goto createCOD_end;

	}



	/* Read number of vertices */

	get1line(fp);

	sscanf(Buffer, "%d", &NumVert);



	/* Allocate storage to save vertices */

	malvert = (struct VERTEX *)malloc(sizeof(struct VERTEX) * NumVert);

	if(!malvert){

		error_no = 4;	/* Vertex buffer malloc error */

		goto createCOD_end;

	}



	/* Read Vertex */

	for(i = 0; i < NumVert; i++){

		vert = malvert + i;

		

		get1line(fp);

		sscanf(Buffer, "%f %f %f", &vert->x, &vert->y, &vert->z);

	}



	/* Close PLY file */

	fclose(fp);



	/* Check number of coordinates */

	NumCoord = NumBone;



	/* Allocate storage to save coordinates */

	malcoord = (struct COORDINATE *)malloc(sizeof(struct COORDINATE) * NumCoord);

	if(!malcoord){

		error_no = 5;	/* Coordinate buffer malloc error */

		goto createCOD_end;

	}



	for(i = 0; i < NumCoord; i++){

		coord = malcoord + i;

		object = (bones + i)->obj;

		matrix = object.tmp;

		start = (bones + i)->restPosition2;

		direction = (bones + i)->direction;

		length = (bones + i)->restLength2;



		/* Check Matrix value */

		for(j = 0; j < 3; j++){

			for(k = 0; k < 3; k++){ 

				coord->m[j][k] = matrix.M[j][k] * (Scalar)RATE;

			}

		}



		/* Check translation amount */

		for(j = 0; j < 3; j++){

			coord->t[j] = matrix.M[j][3] /* * (Scalar)RATE */;

		}



		/* Initialize rotation amount */

		coord->x = coord->y = coord->z = (Scalar)0;



		/* Check index of new coordinate */

		if( object.parent ){

			for(j = 0; j < NumBone; j++){

				if( object.parent == (ObjectPtr)(bones + j)){

					coord->super = j;

				}

			}

		}

		else{

			coord->super = i;

		}



		/* Set center */

		length *= (Scalar)0.5;



		coord->center.x = start.x + length * direction.x;

		coord->center.y = start.y + length * direction.y;

		coord->center.z = start.z + length * direction.z;



		/* Initialize number of assigned vertices */

		coord->numvert = 0;

	}



	/* Check Coordinate of Vertex */

	for(i = 0; i < NumVert; i++){

		vert = malvert + i;



		for(j = 0; j < NumCoord; j++){

			coord = malcoord + j;

			center = (malcoord + j)->center;



			/* Set distance of each Coordinate and Vertex */

			length = (vert->x - center.x) * (vert->x - center.x)

				+ (vert->y - center.y) * (vert->y - center.y)

				+ (vert->z - center.z) * (vert->z - center.z);



			if(j == 0){

				vert->r = length;

				vert->cno = j;

			}

			else{

				if(vert->r > length){

					vert->r = length;

					vert->cno = j;

				}

			}

		}

	}



	/* Count number of assigned vertices for each Coordinate */

	for(i = 0; i < NumCoord; i++){

		coord = malcoord + i;



		for(j = 0; j < NumVert; j++){

			vert = malvert + j;



			if(vert->cno == i){

				coord->numvert++;

			}

		}

	}



	/* Update COD file */

	fp = fopen(codName, "w");

	if(!fp){

		error_no = 6;	/* CODfile fopen error */

		goto createCOD_end;

	}



	fprintf(fp, "#ID\n%s\n", COD_ID);

	fprintf(fp, "#OGP File\nOGP=NULL\n" );

	fprintf(fp, "#Number of Coordinate\n");

	fprintf(fp, "%d\n", NumCoord);



	/* Write Coordinate */

	fprintf(fp, "#Coordinate\n");



	for(i = 0; i < NumCoord; i++){

		coord = malcoord + i;



		/* Write Matrix vluae */

		for(j = 0; j < 3; j++)	

			fprintf(fp, "% 10d % 10d % 10d\n", (int)coord->m[j][0], (int)coord->m[j][1], (int)coord->m[j][2]);



		/* Write translation amount */

		fprintf(fp, "% 10d % 10d % 10d\n", (int)coord->t[0], (int)coord->t[1], (int)coord->t[2]);



		/* Write rotation amount */

		fprintf(fp, "% 10d % 10d % 10d\n", (int)coord->x, (int)coord->y, (int)coord->z);



		/* Write index of new Coordinate */

		fprintf(fp, "%d\n", coord->super);

	}



	/* Write Vertex belonging to each Coordinate */

	fprintf(fp, "#Vertex belong to Coordinate\n");



	for(i = 0; i < NumCoord; i++){

		coord = malcoord + i;



		/* Write Coordinate index */

		fprintf(fp, "%d ", i);



		/* Write number of vertices */

		fprintf(fp, "%d\t", coord->numvert);



		/* Initialize counter */

		counter1 = 0;

		counter2 = 0;



		/* Write Vertex index */

		for(j = 0; j < NumVert; j++){

			vert = malvert + j;



			if(vert->cno == i){

				fprintf(fp, "%d", j);

				counter1++;

				counter2++;



				if(counter1 == NUM_VERTEX && counter2 != coord->numvert){

					fprintf(fp, "\n\t");

					counter1 = 0;

				}

				else if(counter2 == coord->numvert){

					fprintf(fp, "\n");

					counter2 = 0;

				}

				else{

					fprintf(fp, " ");

				}

			}

		}

	}



	/* Close COD file */

	fclose(fp);



createCOD_end:	/* goto jump point */



	free( malvert );

	free( malcoord );

	return error_no;

}



char *get1line(FILE *fp)

{

	int		i = 0;

	char	*s;



	/* Read one line */

	s = fgets(Buffer, BUFFER_LENGTH, fp);



	/* Read the next line in case of comment or whitespace */

	while((Buffer[i] == ' ' || Buffer[i] == '\t' || Buffer[i] == '\n' || Buffer[i] == '#') && s){

		if(Buffer[i] == ' ' || Buffer[i] == '\t'){

			/* Skip whitespace */

			i++;

		}

		else{

			/* Read the next line */

			s = fgets(Buffer, BUFFER_LENGTH, fp);

		}

	}



	return s;

}





