/*			
		Permission to use				 

  Sony Computer Entertainment supplies this lws2rsd software free 	   
  of charge to licensees along with the source code files.		 

  This lws2rsd software is experimental and no support of any kind is 	  
  available.							 

  The licensee is hereby granted permission to use, copy, modify, and	   
  distribute this lws2rsd software and documentation for any purpose. 	  
  Sony Computer Entertainment shall bear no responsibility for any 	  
  damage or loss sustained either directly or indirectly as a result of	   
  using this software. 						   

  Copyright 1997 by Sony Computer Entertainment Inc. All Rights Reserved. 

*/																					

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "io.h"
#include "vecmat.h"
#include "lw.h"
#include "cap.h"
#include "pvt.h"

extern Matrix IdentityMatrix;

#define NEXT_BONE_EXIST	1
#define BONE_GOT_SHADOW_OPTIONS 2

static int		TheModelCount = 0;
static int		TheBoneCount = 0;
static int		TheLocalBoneHead = 0;

static BonePtr	TheBoneArray	= NULL;
static ModelPtr	TheModelArray	= NULL;

static int		TheFirstFrame;
static int		TheLastFrame;

static char ASCII_FLTcode[] = "-Z+X-Y";
static char ASCII_ValidCode[] = "7";
static char ASCII_AngleUnit[] = "1";
static char ASCII_RotationOrder[] = "zxy";

extern int createCOD(char *plyName, int NumBone, BonePtr bones, char *codName);
extern int SaveMOT( char *MOTfileName, int noofBone, BonePtr bone, int totalFrame );


/******************************************************************************/
 void OpenLW( void )
/******************************************************************************/
{
	TheModelCount = 0;
	TheBoneCount = 0;
	TheLocalBoneHead = 0;

	TheBoneArray = (BonePtr)NULL;
	TheModelArray = (ModelPtr)NULL;
}

/******************************************************************************/
 void FreeBones( void )
/******************************************************************************/
{
	BonePtr		bone;
	ObjectPtr	obj;
	int			i;


	bone = TheBoneArray;

	for ( i = 0; i < TheBoneCount; i++ ) {
		obj = &bone[i].obj;

		free( obj->keyframe );
	}

	free( TheBoneArray );
}


/******************************************************************************/
 void FreeModels( void )
/******************************************************************************/
{
	ModelPtr	model;
	ObjectPtr	obj;
	int			i;


	model = TheModelArray;

	for ( i = 0; i < TheModelCount; i++ ) {
		obj = &model[i].obj;

		free( obj->keyframe );
	}

	free( TheModelArray );
}


/******************************************************************************/
 void CloseLW( void )
/******************************************************************************/
{
	if ( TheBoneArray )
		FreeBones();

	if ( TheModelArray )
		FreeModels();
}


/******************************************************************************/
 int GetBoneRestPosition( char *buffer, BonePtr bone )
/******************************************************************************/
{
	Scalar		x, y, z;

	sscanf( buffer, "%f %f %f", &x, &y, &z );

	bone->restPosition.x = x;
	bone->restPosition.y = y;
	bone->restPosition.z = z;
	return 0;
}

/******************************************************************************/
 int GetBoneRestDirection( char *buffer, BonePtr bone )
/******************************************************************************/
{
	Scalar		x, y, z;

	sscanf( buffer, "%f %f %f", &y, &x, &z );	// H, P, B == Y, X, Z

	bone->restDirection.x = x;
	bone->restDirection.y = y;
	bone->restDirection.z = z;
	return 0;
}


/******************************************************************************/
 int GetBoneRestLength( char *buffer, BonePtr bone )
/******************************************************************************/
{
	Scalar		length;

	sscanf( buffer, "%f", &length );
	bone->restLength = length;
	return 0;
}


/******************************************************************************/
 int GetScaleBoneStrength( char *buffer, BonePtr bone )
/******************************************************************************/
{
	int			flag;

	sscanf( buffer, "%d", &flag );
	bone->scaleStrength = flag;
	return 0;
}


/******************************************************************************/
 int GetBoneStrength( char *buffer, BonePtr bone )
/******************************************************************************/
{
	Scalar		strength;

	sscanf( buffer, "%f", &strength );
	bone->strength = strength;
	return 0;
}


/******************************************************************************/
 BonePtr NewBone( void )
/******************************************************************************/
{
	ObjectPtr	obj;


	if ( !TheBoneArray )
		TheBoneArray = (BonePtr)malloc( sizeof(Bone) );
	else
		TheBoneArray = (BonePtr)realloc( TheBoneArray, sizeof(Bone) * ( TheBoneCount+1 ) );

	obj = &TheBoneArray[TheBoneCount].obj;

	obj->parent = NULL;
	obj->id = TheBoneCount;
	obj->parentID = -1;
	obj->keyframe = NULL;

	return &TheBoneArray[TheBoneCount];
}


/******************************************************************************/
 ModelPtr NewModel( void )
/******************************************************************************/
{
	ObjectPtr	obj;


	if ( !TheModelArray )
		TheModelArray = (ModelPtr)malloc( sizeof(Model) );
	else
		TheModelArray = (ModelPtr)realloc( TheModelArray, sizeof(Model) * ( TheModelCount+1 ) );

	obj = &TheModelArray[TheModelCount].obj;

	obj->parent = NULL;
	obj->id = TheModelCount+1;
	obj->parentID = -1;
	obj->keyframe = NULL;

	return &TheModelArray[TheModelCount];
}


/******************************************************************************/
 int GetParentObject( char *buffer, ObjectPtr obj, int offset )
/******************************************************************************/
{
	int		parentID;

	sscanf( buffer, "%d", &parentID );
	obj->parentID = parentID + offset;
	return 0;
}


/******************************************************************************/
 int GetPivotPoint( char *buffer, ModelPtr model )
/******************************************************************************/
{
	Scalar	x, y, z;

	sscanf( buffer, "%f %f %f", &x, &y, &z );

	model->pivot.x = x;
	model->pivot.y = y;
	model->pivot.z = z;
	return 0;
}


/******************************************************************************/
 int GetObjectName( char *buffer, ObjectPtr obj, int type )
/******************************************************************************/
{
	int		i;
	int		backSlash= '\\';
	char	*start, *p, *q;


	if ( type == LW_MODEL ) {
		p = buffer;
		q = obj->PLY;

		while ( *p )
			*q++ = *p++;

		*--q = '\0';	*--q = 'Y';		*--q = 'L';		*--q = 'P';

		if ( !( start = strrchr( buffer, backSlash ) ) )
			return 1;

		start++;
	}
	else {
		start = buffer;
	}


	for (i = 0, p = start; p < &buffer[256]; p++ ) {
		if ( *p == '.' || *p == 0 )
			break;

		obj->name[i] = *p;
		i++;
	}

	obj->name[i] = '\0';
	return 0;
}


/******************************************************************************/
 int GetNumberOfChannel( FILE *fp, char *buffer, int *noofChannel )
/******************************************************************************/
{
	if ( fgets( buffer, 255, fp ) == NULL )
		return 0;

	sscanf( buffer, "%d", noofChannel );
}


/******************************************************************************/
 int GetNumberOfKeyframe( FILE *fp, char *buffer, int *noofKeyframe ) 
/******************************************************************************/
{
	if ( fgets( buffer, 255, fp ) == NULL )
		return 0;

	sscanf( buffer, "%d", noofKeyframe );
	return 0;
}


/******************************************************************************/
 int GetKeyframe9(
	FILE *fp, 
	char *buffer, 
	float *tx, float *ty, float *tz, 
	float *rx, float *ry, float *rz,
	float *sx, float *sy, float *sz, 
	int *frameNo,
	int *linearStatus, 
	float *tension,
	float *bias,
	float *continuity
 )
/******************************************************************************/
{
	if ( fgets( buffer, 255, fp ) == NULL )
		return 0;

	sscanf( buffer, "%f %f %f %f %f %f %f %f %f", tx, ty, tz, rx, ry, rz, sx, sy, sz );

	if ( fgets( buffer, 255, fp ) == NULL )
		return 0;

	sscanf( buffer, "%d %d %f %f %f", frameNo, linearStatus, tension, bias, continuity );
	return 0;
}


/******************************************************************************/
 int GetKeyframes( FILE *fp, char *buffer, ObjectPtr obj )
/******************************************************************************/
{
	KeyframePtr		keyframe = NULL;
	float			sx, sy, sz, rx, ry, rz, tx, ty, tz;
	float			tension, bias, continuity;
	int				frameNo;
	int				linearStatus;
	int				noofChannel;
	int				noofKeyframe;
	int				i;


	GetNumberOfChannel( fp, buffer, &noofChannel );
	GetNumberOfKeyframe( fp, buffer, &noofKeyframe );

	keyframe = (KeyframePtr)malloc( sizeof(Keyframe) * noofKeyframe );
	if ( !keyframe )
		return -1;

	for ( i = 0; i < noofKeyframe; i++ ) {
		GetKeyframe9( fp, buffer, 
			&tx, &ty, &tz, &ry, &rx, &rz, &sx, &sy, &sz, 
			&frameNo, &linearStatus, &tension, &bias, &continuity
		);

		keyframe[i].frameNo = frameNo;
		keyframe[i].linearStatus = linearStatus;

		keyframe[i].s.x = sx;		keyframe[i].s.y = sy;		keyframe[i].s.z = sz;
		keyframe[i].r.x = rx;		keyframe[i].r.y = ry;		keyframe[i].r.z = rz;
		keyframe[i].t.x = tx;		keyframe[i].t.y = ty;		keyframe[i].t.z = tz;
	}

	obj->noofKeyframe = noofKeyframe;
	obj->keyframe = keyframe;
}


/******************************************************************************/
 int LoadBone( FILE *fp, char *buffer, BonePtr bone )
/******************************************************************************/
{
	char token[64];
	char *next;


	while ( 1 ) {
		if ( fgets( buffer, 255, fp ) == NULL )
			break;
		
		next = GetToken( buffer, token );

		if ( !strcmp( token, "BoneName" ) ) {
			GetObjectName( next, &bone->obj, LW_BONE );
			continue;
		}

		if ( !strcmp( token, "ShowBone" ) ) {
			continue;
		}

		if ( !strcmp( token, "BoneActive" ) ) {
			continue;
		}

		if ( !strcmp( token, "BoneRestPosition" ) ) {
			GetBoneRestPosition( next, bone );
			continue;
		}

		if ( !strcmp( token, "BoneRestDirection" ) ) {
			GetBoneRestDirection( next, bone );
			continue;
		}

		if ( !strcmp( token, "BoneRestLength" ) ) {
			GetBoneRestLength( next, bone );
			continue;
		}

		if ( !strcmp( token, "ScaleBoneStrength" ) ) {
			GetScaleBoneStrength( next, bone );
			continue;
		}

		if ( !strcmp( token, "BoneStrength" ) ) {
			GetBoneStrength( next, bone );
			continue;
		}


		if ( !strcmp( token, "BoneMotion" ) ) {
			GetKeyframes( fp, buffer, &bone->obj );
			continue;
		}

		if ( !strcmp( token, "EndBehavior" ) ) {
			continue;
		}

		if ( !strcmp( token, "ParentObject" ) ) {
			GetParentObject( next, &bone->obj, TheLocalBoneHead-1 );
			continue;
		}

		if ( !strcmp( token, "AddBone" ) ) {
			return NEXT_BONE_EXIST;
		}

		if ( !strcmp( token, "ShadowOptions" ) )
			return BONE_GOT_SHADOW_OPTIONS;
	}

	return 0;
}


/******************************************************************************/
 void SetupBoneHierarchy( void )
/******************************************************************************/
 {
	BonePtr		ptr, head, tail;


	head = &TheBoneArray[TheLocalBoneHead];
	tail = &TheBoneArray[TheBoneCount-1];


	for ( ptr = head; ptr <= tail; ptr++ ) {
		if ( ptr->obj.parentID >= 0 )
			ptr->obj.parent = (ObjectPtr)&TheBoneArray[ptr->obj.parentID];
	}
 }


/******************************************************************************/
 int LoadModel( FILE *fp, char *buffer, ModelPtr model )
/******************************************************************************/
{
	int			rtv;
	char		token[64];
	char		*next = NULL;
	BonePtr		bone = NULL;

	while ( 1 ) {
		if ( fgets( buffer, 255, fp ) == NULL )
			break;
		
		next = GetToken( buffer, token );

		if ( !strcmp( token, "ShowObject" ) ) {
			continue;
		}

		if ( !strcmp( token, "ObjectMotion" ) ) {
			GetKeyframes( fp, next, &model->obj );
			continue;
		}

		if ( !strcmp( token, "EndBehavior" ) ) {
			continue;
		}

		if ( !strcmp( token, "PivotPoint" ) ) {
			GetPivotPoint( next, model );
			continue;
		}

		if ( !strcmp( token, "ParentObject" ) ) {
			GetParentObject( next, &model->obj, 0 );
			continue;
		}

		if ( !strcmp( token, "AddBone" ) ) {
			TheLocalBoneHead = TheBoneCount;

			do {
				if ( !( bone = NewBone() ) )
					return 1;

				rtv = LoadBone( fp, buffer, bone );
				TheBoneCount++;

				if ( rtv == BONE_GOT_SHADOW_OPTIONS )
					break; // this must be final line of an object

			} while ( rtv == NEXT_BONE_EXIST );

			SetupBoneHierarchy();
			continue;
		}

		if ( !strcmp( token, "ShadowOptions" ) )
			break; // this must be final line of an object
	}

	return 0;
}


/******************************************************************************/
 void SetupModelHierarchy( void )
/******************************************************************************/
 {
	ModelPtr		ptr, head, tail;


	head = &TheModelArray[0];
	tail = &TheModelArray[TheModelCount-1];


	for ( ptr = head; ptr <= tail; ptr++ ) {
		if ( ptr->obj.parentID >= 0 )
			ptr->obj.parent = (ObjectPtr)&TheModelArray[ptr->obj.parentID];
	}
 }


/******************************************************************************/
 void LW2PS( Scalar scaleLw2Ps )	// one model, multiple bones
/******************************************************************************/
{
	int			i, j;
	ObjectPtr	obj;
	ModelPtr	model;
	BonePtr		bone;
	KeyframePtr	key;
	Param		s, r, t, s2, r2, t2;
	Matrix		mat;
	char		fileName[64];
	char		*p, *q;
	int			rtv;
	Vector3		ez = { (Scalar)0.0, (Scalar)0.0, (Scalar)1.0 };



	SetFLTcode( ASCII_FLTcode );
	SetValidCode( ASCII_ValidCode );
	SetRotationOrder( ASCII_RotationOrder );
	SetAngleUnit( ASCII_AngleUnit );

	// Model
	for ( i = 0; i < TheModelCount; i++ ) {
		model = &TheModelArray[i];
		s.x = s.y = s.z = (Scalar)1.0;
		r.x = r.y = r.z = (Scalar)0.0;
		t = model->pivot;
		Transform( scaleLw2Ps, &s, &r, &t, &s2, &r2, &model->pivot2 );

		// keyframe parameters
		obj = &model->obj;

		for ( j = 0; j < obj->noofKeyframe; j++ ) {
			key = &obj->keyframe[j];
			Transform( scaleLw2Ps, &key->s, &key->r, &key->t, &key->s2, &key->r2, &key->t2 );
		}
	}

	// Bone
	for ( i = 0; i < TheBoneCount; i++ ) {
		bone = &TheBoneArray[i];
		obj = &bone->obj;

		bone->radius = (Scalar)(bone->restLength / PI2);

		bone->restLength2 = bone->restLength * scaleLw2Ps;
		bone->scaleStrength2 = bone->scaleStrength;
		bone->strength2 = bone->strength;

		s.x = s.y = s.z = (Scalar)1.0;
		r = bone->restDirection;
		t = bone->restPosition;
		Transform( scaleLw2Ps, &s, &r, &t, &s2, &bone->restDirection2, &bone->restPosition2 );

		r = bone->restDirection2;
		t = bone->restPosition2; 
		SetPSXMatrix( &s, &r, &t, &obj->mat );
	}

	//		concatenate matrices
	for ( i = 0; i < TheBoneCount; i++ ) {
		bone = &TheBoneArray[i];		
		obj = &bone->obj;
		mat = obj->mat;

		while ( obj ) {
			if ( obj->parent )
				mat = MatrixMult( &obj->parent->mat, &mat );

			obj = obj->parent;
		}

		bone->obj.tmp = mat;
	}

	//		map PS local to PS world
	for ( i = 0; i < TheBoneCount; i++ ) {
		bone = &TheBoneArray[i];		
		obj = &bone->obj;
		mat = obj->tmp;

		DecomposeTRS2t_r_s( &mat, &t2, &r2, &s2 );
		bone->restDirection2 = r2;
		bone->restPosition2 = t2;

		mat = SetZYXRotationMatrix( &r2 );
		bone->direction = MatrixVector43( &mat, &ez );
	}


	for ( i = 0; i < TheBoneCount; i++ ) {
		Matrix		mat2, mat3;

		bone = &TheBoneArray[i];		
		obj = &bone->obj;
	
		mat = IdentityMatrix;

		while ( obj ) {
			if ( obj->parent ) {
				mat2 = obj->parent->mat;

				mat2.M[0][3] = (Scalar)0.0;
				mat2.M[1][3] = (Scalar)0.0;
				mat2.M[2][3] = (Scalar)0.0;

				mat = MatrixMult( &mat2, &mat );
			}

			obj = obj->parent;
		}

		t.x = bone->obj.mat.M[0][3];
		t.y = bone->obj.mat.M[1][3];
		t.z = bone->obj.mat.M[2][3];

		t2  = MatrixVector43( &mat, &t );
		bone->obj.tmp = SetTranslationMatrix( &t2 );

		// keyframe parameters
		for ( j = 0; j < bone->obj.noofKeyframe; j++ ) {
			key = &(bone->obj.keyframe[j]);
			Transform( scaleLw2Ps, &key->s, &key->r, &key->t, &key->s2, &key->r2, &key->t2 );

			SetPSXMatrix( &key->s2, &key->r2, &key->t2, &mat2 );
			mat2 = MatrixMult( &mat, &mat2 );
			DecomposeTRS2t_r_s( &mat2, &key->t2, &key->r2, &key->s2 );

			mat3 = SetZYXRotationMatrix( &bone->restDirection2 );
			InverseMatrix( &mat3, &mat3 );
			mat2 = SetZYXRotationMatrix( &key->r2 );
			mat2 = MatrixMult( &mat2, &mat3 );
			key->r2 = DecomposeRotationMatrixZYX( &mat2 );
		}
	}


	// COD file
	p = TheModelArray[0].obj.PLY;
	q = fileName;

	while ( *p )
		*q++ = *p++;

	*q-- = '\0';	*q-- = 'D';		*q-- = 'O';		*q-- = 'C';

	rtv = createCOD( TheModelArray[0].obj.PLY, TheBoneCount, TheBoneArray, fileName );

	if ( rtv ) {
		CloseLW();
		fprintf( stderr, "Some error %d occured while COD file creating.\n", rtv );
		exit( 1 );
	}

	// PVT file

	p = TheModelArray[0].obj.PLY;
	q = fileName;

	while ( *p )
		*q++ = *p++;

	*q-- = '\0';	*q-- = 'T';		*q-- = 'V';		*q-- = 'P';

	rtv = SavePVT( fileName, TheBoneCount, (ObjectPtr)TheBoneArray, LW_BONE );

	if ( rtv ) {
		CloseLW();
		fprintf( stderr, "Some error %d occured while PVT file creating.\n", rtv );
		exit( 1 );
	}

	// MOT file

	p = TheModelArray[0].obj.PLY;
	q = fileName;

	while ( *p )
		*q++ = *p++;

	*q-- = '\0';	*q-- = 'T';		*q-- = 'O';		*q-- = 'M';

	rtv = SaveMOT( fileName, TheBoneCount, TheBoneArray, TheLastFrame-TheFirstFrame );

	if ( rtv ) {
		CloseLW();
		fprintf( stderr, "Some error %d occured while MOT file creating.\n", rtv );
		exit( 1 );
	}
}


/******************************************************************************/
 int MasterParser( FILE *fp, char *buffer )
/******************************************************************************/
{
	char		token[64];
	char		*next = NULL;
	ModelPtr	model = NULL;


	while ( 1 ) {
		if ( fgets( buffer, 255, fp ) == NULL )
			break;

		next = GetToken( buffer, token );

		if ( !strcmp( token, "FirstFrame" ) ) {
			sscanf( next, "%d", &TheFirstFrame );
			continue;
		}

		if ( !strcmp( token, "LastFrame" ) ) {
			sscanf( next, "%d", &TheLastFrame );
			continue;
		}
		
		if ( !strcmp( token, "LoadObject" ) ) {

			if ( !( model = NewModel() ) )
				return 1;
			
			GetObjectName( next, &model->obj, LW_MODEL );
			LoadModel( fp, buffer, model );
			TheModelCount++;

			continue;
		}

		if ( !strcmp( token, "AddNullObject" ) ) {

			if ( !( model = NewModel() ) )
				return 1;

			GetObjectName( buffer, &model->obj, LW_MODEL );
			LoadModel( fp, buffer, model );
			TheModelCount++;

			continue;
		}
	}

	SetupModelHierarchy();
}
