/*			
		Permission to use				 

  Sony Computer Entertainment supplies this lws2rsd software free 	   
  of charge to licensees along with the source code files.		 

  This lws2rsd software is experimental and no support of any kind is 	  
  available.							 

  The licensee is hereby granted permission to use, copy, modify, and	   
  distribute this lws2rsd software and documentation for any purpose. 	  
  Sony Computer Entertainment shall bear no responsibility for any 	  
  damage or loss sustained either directly or indirectly as a result of	   
  using this software. 						   

  Copyright 1997 by Sony Computer Entertainment Inc. All Rights Reserved. 

*/																					

#define VEC_MAT
#include <math.h>

#include "myTypes.h"
#include "vecmat.h"

Matrix ZeroMatrix = 
{
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)1.0
};

Matrix IdentityMatrix = 
{
	(Scalar)1.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)1.0,	(Scalar)0.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)1.0,	(Scalar)0.0,
	(Scalar)0.0,	(Scalar)0.0,	(Scalar)0.0,	(Scalar)1.0
};

#define MIN_SCALE (Scalar)(1.0/4096.0)
#define MAX_SCALE (Scalar)(8 - 1.0/4096.0)


/******************************************************************************/
 Param DEGtoRAD( ParamPtr angle )
/******************************************************************************/
{
	Param theta;

	theta.x = (Scalar)( angle->x * RAD );
	theta.y = (Scalar)( angle->y * RAD );
	theta.z = (Scalar)( angle->z * RAD );

	return theta;
}

/******************************************************************************/
 Param RADtoDEG( ParamPtr theta )
/******************************************************************************/
{
	Param angle;

	angle.x = (Scalar)( theta->x / RAD );
	angle.y = (Scalar)( theta->y / RAD );
	angle.z = (Scalar)( theta->z / RAD );

	return angle;
}


/******************************************************************************/
 Vector3 MatrixVector43( MatrixPtr m, Vector3Ptr v )
/******************************************************************************/
{
	Vector3 	MatVec;

	MatVec.x = m->M[0][0] * v->x + m->M[0][1] * v->y + m->M[0][2] * v->z + m->M[0][3];
	MatVec.y = m->M[1][0] * v->x + m->M[1][1] * v->y + m->M[1][2] * v->z + m->M[1][3];
	MatVec.z = m->M[2][0] * v->x + m->M[2][1] * v->y + m->M[2][2] * v->z + m->M[2][3];

	return ( MatVec );
}


/******************************************************************************/
 Vector MatrixVector( MatrixPtr m, VectorPtr v )
/******************************************************************************/
{
	Vector 	MatVec;

	MatVec.x = m->M[0][0] * v->x + m->M[0][1] * v->y + m->M[0][2] * v->z + m->M[0][3] * v->w;
	MatVec.y = m->M[1][0] * v->x + m->M[1][1] * v->y + m->M[1][2] * v->z + m->M[1][3] * v->w;
	MatVec.z = m->M[2][0] * v->x + m->M[2][1] * v->y + m->M[2][2] * v->z + m->M[2][3] * v->w;
	MatVec.w = m->M[3][0] * v->x + m->M[3][1] * v->y + m->M[3][2] * v->z + m->M[3][3] * v->w;

	return ( MatVec );
}


/******************************************************************************/
 Matrix TransposeMatrix( MatrixPtr A )
/******************************************************************************/
{
  Matrix At;

  At.M[0][0] = A->M[0][0]; At.M[0][1] = A->M[1][0]; At.M[0][2] = A->M[2][0]; At.M[0][3] = A->M[3][0];
  At.M[1][0] = A->M[0][1]; At.M[1][1] = A->M[1][1]; At.M[1][2] = A->M[2][1]; At.M[1][3] = A->M[3][1];
  At.M[2][0] = A->M[0][2]; At.M[2][1] = A->M[1][2]; At.M[2][2] = A->M[2][2]; At.M[2][3] = A->M[3][2];
  At.M[3][0] = A->M[0][3]; At.M[3][1] = A->M[1][3]; At.M[3][2] = A->M[2][3]; At.M[3][3] = A->M[3][3];

  return ( At );
}


/******************************************************************************/
 Scalar DeterminantMatrix( MatrixPtr A )
/******************************************************************************/
{
  Scalar	 detA;

  detA =  ( A->M[0][0] * A->M[1][1] - A->M[0][1] * A->M[1][0] ) * ( A->M[2][2] * A->M[3][3] - A->M[2][3] * A->M[3][2] )
		+ ( A->M[0][0] * A->M[1][2] - A->M[0][2] * A->M[1][0] ) * ( A->M[2][3] * A->M[3][1] - A->M[2][1] * A->M[3][3] )
		+ ( A->M[0][0] * A->M[1][3] - A->M[0][3] * A->M[1][0] ) * ( A->M[2][1] * A->M[3][2] - A->M[2][2] * A->M[3][1] )
		+ ( A->M[2][0] * A->M[3][3] - A->M[2][3] * A->M[3][0] ) * ( A->M[0][1] * A->M[1][2] - A->M[0][2] * A->M[1][1] )
		+ ( A->M[2][0] * A->M[3][1] - A->M[2][1] * A->M[3][0] ) * ( A->M[0][2] * A->M[1][3] - A->M[0][3] * A->M[1][2] )
		+ ( A->M[2][0] * A->M[3][2] - A->M[2][2] * A->M[3][0] ) * ( A->M[0][3] * A->M[1][1] - A->M[0][1] * A->M[1][3] );

  return ( detA );
}


/******************************************************************************/
 int InverseMatrix( MatrixPtr A, MatrixPtr IA )
/******************************************************************************/
{
  Scalar		 detA;
  Scalar		 a, b, c, d;
  Scalar		 e, f, g, h;
  Scalar		 i, j, k, l;
  Scalar		 m, n, o, p;
  Scalar		 d1, d2, d3, d4, d5, d6;
  Scalar		 d7, d8, d9, dA, dB, dC;

  a = A->M[0][0];   b = A->M[0][1]; c = A->M[0][2]; d = A->M[0][3];
  e = A->M[1][0];   f = A->M[1][1]; g = A->M[1][2]; h = A->M[1][3];
  i = A->M[2][0];   j = A->M[2][1]; k = A->M[2][2]; l = A->M[2][3];
  m = A->M[3][0];   n = A->M[3][1]; o = A->M[3][2]; p = A->M[3][3];

  d1 = b*e - a*f;   d2 = l*o - k*p; d3 = c*f - b*g; d4 = l*m - i*p;
  d5 = c*h - d*g;   d6 = i*n - j*m; d7 = a*h - d*e; d8 = j*o - k*n;
  d9 = a*g - c*e;   dA = l*n - j*p; dB = d*f - b*h; dC = i*o - k*m;

  detA = d1*d2 + d3*d4 + d5*d6 + d7*d8 + d9*dA + dB*dC;

  if ( detA == 0.0 )
	   return ( -1 );

  detA = 1 / detA;

  IA->M[0][0] = ( h * d8 - f * d2 + g * dA ) * detA;
  IA->M[1][0] = ( e * d2 - g * d4 - h * dC ) * detA;
  IA->M[2][0] = ( h * d6 - e * dA + f * d4 ) * detA;
  IA->M[3][0] = ( f * dC - g * d6 - e * d8 ) * detA;

  IA->M[0][1] = ( b * d2 - c * dA - d * d8 ) * detA;
  IA->M[1][1] = ( c * d4 + d * dC - a * d2 ) * detA;
  IA->M[2][1] = ( a * dA - b * d4 - d * d6 ) * detA;
  IA->M[3][1] = ( a * d8 - b * dC + c * d6 ) * detA;

  IA->M[0][2] = ( n * d5 + o * dB - p * d3 ) * detA;
  IA->M[1][2] = ( o * d7 - p * d9 - m * d5 ) * detA;
  IA->M[2][2] = ( - n * d7 - p * d1 - m * dB ) * detA;
  IA->M[3][2] = ( m * d3 + n * d9 + o * d1 ) * detA;

  IA->M[0][3] = ( l * d3 - j * d5 - k * dB ) * detA;
  IA->M[1][3] = ( i * d5 + l * d9 - k * d7 ) * detA;
  IA->M[2][3] = ( i * dB + j * d7 + l * d1 ) * detA;
  IA->M[3][3] = ( - k * d1 - i * d3 - j * d9 ) * detA;

  return ( OK );
}


/******************************************************************************/
 Matrix MatrixMult( MatrixPtr B, MatrixPtr A )
/******************************************************************************/
{
	Matrix BAsafe;

	BAsafe.M[0][0]=B->M[0][0]* A->M[0][0]+B->M[0][1]* A->M[1][0]+B->M[0][2]* A->M[2][0]+B->M[0][3]* A->M[3][0];
	BAsafe.M[0][1]=B->M[0][0]* A->M[0][1]+B->M[0][1]* A->M[1][1]+B->M[0][2]* A->M[2][1]+B->M[0][3]* A->M[3][1];
	BAsafe.M[0][2]=B->M[0][0]* A->M[0][2]+B->M[0][1]* A->M[1][2]+B->M[0][2]* A->M[2][2]+B->M[0][3]* A->M[3][2];
	BAsafe.M[0][3]=B->M[0][0]* A->M[0][3]+B->M[0][1]* A->M[1][3]+B->M[0][2]* A->M[2][3]+B->M[0][3]* A->M[3][3];

	BAsafe.M[1][0]=B->M[1][0]* A->M[0][0]+B->M[1][1]* A->M[1][0]+B->M[1][2]* A->M[2][0]+B->M[1][3]* A->M[3][0];
	BAsafe.M[1][1]=B->M[1][0]* A->M[0][1]+B->M[1][1]* A->M[1][1]+B->M[1][2]* A->M[2][1]+B->M[1][3]* A->M[3][1];
	BAsafe.M[1][2]=B->M[1][0]* A->M[0][2]+B->M[1][1]* A->M[1][2]+B->M[1][2]* A->M[2][2]+B->M[1][3]* A->M[3][2];
	BAsafe.M[1][3]=B->M[1][0]* A->M[0][3]+B->M[1][1]* A->M[1][3]+B->M[1][2]* A->M[2][3]+B->M[1][3]* A->M[3][3];

	BAsafe.M[2][0]=B->M[2][0]* A->M[0][0]+B->M[2][1]* A->M[1][0]+B->M[2][2]* A->M[2][0]+B->M[2][3]* A->M[3][0];
	BAsafe.M[2][1]=B->M[2][0]* A->M[0][1]+B->M[2][1]* A->M[1][1]+B->M[2][2]* A->M[2][1]+B->M[2][3]* A->M[3][1];
	BAsafe.M[2][2]=B->M[2][0]* A->M[0][2]+B->M[2][1]* A->M[1][2]+B->M[2][2]* A->M[2][2]+B->M[2][3]* A->M[3][2];
	BAsafe.M[2][3]=B->M[2][0]* A->M[0][3]+B->M[2][1]* A->M[1][3]+B->M[2][2]* A->M[2][3]+B->M[2][3]* A->M[3][3];

	BAsafe.M[3][0]=B->M[3][0]* A->M[0][0]+B->M[3][1]* A->M[1][0]+B->M[3][2]* A->M[2][0]+B->M[3][3]* A->M[3][0];
	BAsafe.M[3][1]=B->M[3][0]* A->M[0][1]+B->M[3][1]* A->M[1][1]+B->M[3][2]* A->M[2][1]+B->M[3][3]* A->M[3][1];
	BAsafe.M[3][2]=B->M[3][0]* A->M[0][2]+B->M[3][1]* A->M[1][2]+B->M[3][2]* A->M[2][2]+B->M[3][3]* A->M[3][2];
	BAsafe.M[3][3]=B->M[3][0]* A->M[0][3]+B->M[3][1]* A->M[1][3]+B->M[3][2]* A->M[2][3]+B->M[3][3]* A->M[3][3];

	return ( BAsafe );
}


/******************************************************************************/
 Matrix SetScalingMatrix( ParamPtr s )
/******************************************************************************/
{
	Matrix Scale = IdentityMatrix;
	
	Scale.M[0][0] = s->x;
	Scale.M[1][1] = s->y;
	Scale.M[2][2] = s->z;

	return ( Scale );
}

/******************************************************************************/
 Matrix SetTranslationMatrix( ParamPtr t )
/******************************************************************************/
{
	Matrix Trans = IdentityMatrix;

	Trans.M[0][3] = t->x;
	Trans.M[1][3] = t->y;
	Trans.M[2][3] = t->z;

	return ( Trans );
}


/******************************************************************************/
 Matrix SetXRotationMatrix( Scalar rx )
/******************************************************************************/
{
	Matrix Rot = IdentityMatrix;
	Scalar Sx = (Scalar)sin( rx );
	Scalar Cx = (Scalar)cos( rx );
	
	Rot.M[1][1] = Cx;	Rot.M[1][2] = -Sx;
	Rot.M[2][1] = Sx;	Rot.M[2][2] = Cx;
	return Rot;
}


/******************************************************************************/
 Matrix SetYRotationMatrix( Scalar ry )
/******************************************************************************/
{
	Matrix Rot = IdentityMatrix;
	Scalar Sy = (Scalar)sin( ry );
	Scalar Cy = (Scalar)cos( ry );
	
	Rot.M[0][0] = Cy;	Rot.M[0][2] = Sy;
	Rot.M[2][0] =-Sy;	Rot.M[2][2] = Cy;  
	return Rot;
}


/******************************************************************************/
 Matrix SetZRotationMatrix( Scalar rz )
/******************************************************************************/
{
	Matrix Rot = IdentityMatrix;
	Scalar Sz = (Scalar)sin( rz );
	Scalar Cz = (Scalar)cos( rz );
	
	Rot.M[0][0] = Cz;	Rot.M[0][1] =-Sz;
	Rot.M[1][0] = Sz;	Rot.M[1][1] = Cz;
	return Rot;
}


/******************************************************************************/
 Matrix SetZYXRotationMatrix( ParamPtr r )
// rx, ry, and rz should be RADIAN !!
/******************************************************************************/
{
	Scalar	Sx = (Scalar)sin( r->x ); 
	Scalar	Cx = (Scalar)cos( r->x );
	Scalar	Sy = (Scalar)sin( r->y );
	Scalar	Cy = (Scalar)cos( r->y );
	Scalar	Sz = (Scalar)sin( r->z );
	Scalar	Cz = (Scalar)cos( r->z );
	Matrix	Rot = IdentityMatrix;
  
	Rot.M[0][0] =Cy*Cz;				Rot.M[0][1] =-Cy*Sz; 			Rot.M[0][2] =Sy;
	Rot.M[1][0] =Cx*Sz+Sx*Sy*Cz;	Rot.M[1][1] =Cx*Cz-Sx*Sy*Sz; 	Rot.M[1][2] =-Sx*Cy;
	Rot.M[2][0] =Sx*Sz-Cx*Sy*Cz; 	Rot.M[2][1] =Sx*Cz+Cx*Sy*Sz;	Rot.M[2][2] =Cx*Cy;

	return ( Rot );
}


/******************************************************************************/
  Matrix SetupMatrix( ParamPtr s, ParamPtr r, ParamPtr t )
/******************************************************************************/
{
	Matrix	S;
	Matrix	Rx, Ry, Rz, Rzxy;
	Matrix	T;
	Matrix	TRS;

	S	= SetScalingMatrix( s );
	
	Rx	= SetXRotationMatrix( r->x );
	Ry	= SetYRotationMatrix( r->y );
	Rz	= SetZRotationMatrix( r->z );
	Rzxy= MatrixMult( &Rx, &Rz );
	Rzxy= MatrixMult( &Ry, &Rzxy );	// light wave

	T	= SetTranslationMatrix( t );

	TRS = MatrixMult( &Rzxy, &S );
	TRS	= MatrixMult( &T, &TRS );

	return TRS;
}


/******************************************************************************/
 Param  DecomposeScalingMatrix( MatrixPtr S )
/******************************************************************************/
{
	Param s;

	s.x = S->M[0][0];
	s.y = S->M[1][1];
	s.z = S->M[2][2];

	return s;
}
	

/******************************************************************************/
 Param DecomposeTranslationMatrix( MatrixPtr T )
/******************************************************************************/
{
	Param t;

	t.x = T->M[0][3];
	t.y = T->M[1][3];
	t.z = T->M[2][3];

	return t;
}


/******************************************************************************/ 
 Param  DecomposeRotationMatrixZYX( MatrixPtr Rot )
 // General rotation matrix is decomposed to RxRyRz rotation matrix angles.
/******************************************************************************/
{
  Scalar	 SinX, CosX, sinX, cosX;
  Param		 theta, fi;

  theta.x = (Scalar)(-atan2( Rot->M[1][2], Rot->M[2][2] ) );
  fi.x = ( ( theta.x <= 0 ) ? (Scalar)( theta.x + PI ) : (Scalar)( theta.x - PI ) );

  SinX = (Scalar)sin(theta.x);	  	CosX = (Scalar)cos(theta.x);
  sinX = (Scalar)sin( fi.x );		cosX = (Scalar)cos(fi.x);
  
  theta.y	= (Scalar)atan2( Rot->M[0][2], Rot->M[2][2] * CosX - Rot->M[1][2] * SinX );
  fi.y		= (Scalar)atan2( Rot->M[0][2], Rot->M[2][2] * cosX - Rot->M[1][2] * sinX );
 
  theta.z	= (Scalar)atan2( Rot->M[1][0] * CosX + Rot->M[2][0] * SinX, Rot->M[1][1] * CosX + Rot->M[2][1] * SinX );
  fi.z		= (Scalar)atan2( Rot->M[1][0] * cosX + Rot->M[2][0] * sinX, Rot->M[1][1] * cosX + Rot->M[2][1] * sinX );
  
  SinX = ABS(theta.x) + ABS(theta.y) + ABS(theta.z);
  CosX = ABS(fi.x) + ABS(fi.y) + ABS(fi.z);
  
  if ( SinX < CosX )
	return theta;
  else
	return fi;
}



/******************************************************************************/
 int DecomposeTRStoT_R_S
( 
 	MatrixPtr TRS,			// input concatenated matrix
 	MatrixPtr T,			// output translation matrix
 	MatrixPtr R, 			// output rotation matrix
 	MatrixPtr S				// output scalling matrix
)
/******************************************************************************/
{
	int			i;
    Scalar		rs[3];	// rs[i] = 1.0 / s[i]	( i = 0, 1, 2 )


	// check null pointer
	if ( !TRS || !T || !R || !S )
		return NG;

	// clean up output matrices
	*T = *R = *S = IdentityMatrix;
	
	// extract translation parameters
	T->M[0][3] = TRS->M[0][3];
	T->M[1][3] = TRS->M[1][3];
	T->M[2][3] = TRS->M[2][3];

	//  extract scaling factors & set up scaling matrix
	for ( i = 0; i < 3; i++ ) {
		// Scaling factor is allways plus value for PlayStation.
		S->M[i][i] = (Scalar)sqrt( TRS->M[0][i] * TRS->M[0][i] + TRS->M[1][i] * TRS->M[1][i] + TRS->M[2][i] * TRS->M[2][i] );
		
		if ( S->M[i][i] < MIN_SCALE )
			return NG;					// no rotation no scaling!
		else
			rs[i] = (Scalar)( 1.0 / S->M[i][i] );
	}

	// set up rotation matrix
	for ( i = 0; i < 3; i++ ) {
		R->M[0][i] = TRS->M[0][i] * rs[i];
		R->M[1][i] = TRS->M[1][i] * rs[i];
		R->M[2][i] = TRS->M[2][i] * rs[i];
	}

	return OK;
}


/******************************************************************************/
 int  DecomposeTRS2t_r_s
( 
 	MatrixPtr 	TRS, 		// input concatenated matrix
 	ParamPtr 	t,			// translation parameters
 	ParamPtr 	r,			// rotation parameters in radian,  rotation order is z y x. 
 	ParamPtr 	s			// scaling parametaers
)
/******************************************************************************/
{
	int			j;
    Scalar		ss[3], rs[3];	// S[j][j] = ss[j], rs[j] = 1.0 / s[j]	( j = 0, 1, 2 )
	Matrix		R;				// temporary Rotation matrix


	// check NULL pointer
	if ( !TRS || !t || !r || !s )
	  return NG;

	// extract translation parameters	
	t->x = TRS->M[0][3];
	t->y = TRS->M[1][3];
	t->z = TRS->M[2][3];

	// extract scaling factors
	for ( j = 0; j < 3; j++ ) {
		ss[j] = (Scalar)sqrt( TRS->M[0][j] * TRS->M[0][j] + TRS->M[1][j] * TRS->M[1][j] + TRS->M[2][j] * TRS->M[2][j] );
		
		if ( ss[j] < MIN_SCALE )
			return NG;					// no rotation no scaling!
		else
			rs[j] = (Scalar)( 1.0 / ss[j] );
	}

	// set up rotation matrix
	R = IdentityMatrix;
	
	for ( j = 0; j < 3; j++ ) {
		R.M[0][j] = TRS->M[0][j] * rs[j];
		R.M[1][j] = TRS->M[1][j] * rs[j];
		R.M[2][j] = TRS->M[2][j] * rs[j];
	}

    *r = DecomposeRotationMatrixZYX( &R );

	// set up scaling paramters
	s->x = ss[0];
	s->y = ss[1];
	s->z = ss[2];

	return OK;
}


/******************************************************************************/
 int  ComposeTRSfromT_R_S
( 
 	MatrixPtr TRS,			// composite matrix
 	MatrixPtr T,			// translation matrix
 	MatrixPtr R,			// rotation matrix
 	MatrixPtr S				// scaling matrix
)
/******************************************************************************/
{
	int		i, j;


	if ( !TRS || !T || !R || !S )
		return NG;

	// setup scaling and rotaiton part
	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ )
			TRS->M[i][j] = S->M[j][j] * R->M[i][j];

	// setup translation part
	TRS->M[0][3] = T->M[0][3];
	TRS->M[1][3] = T->M[1][3];
	TRS->M[2][3] = T->M[2][3];

	// clean up projection part
	TRS->M[3][0] = TRS->M[3][1] = TRS->M[3][2] = (Scalar)0.0; TRS->M[3][3] = (Scalar)1.0;

	return OK;
}


/******************************************************************************/
 Matrix  SetInverseTranslationMatrix( ParamPtr t )
/******************************************************************************/
{
	Matrix	Itrans = IdentityMatrix;

	if ( !t )
		return Itrans;

	Itrans.M[0][3] = -t->x;
	Itrans.M[1][3] = -t->y;
	Itrans.M[2][3] = -t->z;
	
	return Itrans;
}


/******************************************************************************/
 Matrix  SetInverseRotationMatrix( ParamPtr r )
/******************************************************************************/
{   
    Param 	ir;

	ir.x = -r->x;
	ir.y = -r->y;
	ir.z = -r->z;

	return SetZYXRotationMatrix( &ir );
}


/******************************************************************************/
 Matrix SetInverseScalingMatrix( ParamPtr s )
/******************************************************************************/
{   
	Matrix Iscale = IdentityMatrix;

	if ( !s )
		return Iscale;
	
	Iscale.M[0][0] = (Scalar)( 1.0 / s->x );
	Iscale.M[1][1] = (Scalar)( 1.0 / s->y );
	Iscale.M[2][2] = (Scalar)( 1.0 / s->z );

	return Iscale;
}


/******************************************************************************/
 void SetPSXMatrix( ParamPtr s, ParamPtr r, ParamPtr t, MatrixPtr TRS )
/******************************************************************************/
{
  static Scalar xS, yS, zS;
  static Scalar	Sx, Sy, Sz;
  static Scalar	Cx, Cy, Cz;

  Sx = (Scalar)sin( r->x ); Cx = (Scalar)cos( r->x );
  Sy = (Scalar)sin( r->y ); Cy = (Scalar)cos( r->y );
  Sz = (Scalar)sin( r->z ); Cz = (Scalar)cos( r->z );

  xS = s->x < MIN_SCALE ? MIN_SCALE : s->x > MAX_SCALE ? MAX_SCALE : s->x;
  yS = s->y < MIN_SCALE ? MIN_SCALE : s->y > MAX_SCALE ? MAX_SCALE : s->y;
  zS = s->z < MIN_SCALE ? MIN_SCALE : s->z > MAX_SCALE ? MAX_SCALE : s->z;
  
  TRS->M[0][0] =xS*Cy*Cz;			 TRS->M[0][1] =-yS*Cy*Sz; 		    TRS->M[0][2] =zS*Sy;		TRS->M[0][3] =t->x;
  TRS->M[1][0] =xS*(Cx*Sz+Sx*Sy*Cz); TRS->M[1][1] =yS*(Cx*Cz-Sx*Sy*Sz); TRS->M[1][2] =-zS*Sx*Cy;	TRS->M[1][3] =t->y;
  TRS->M[2][0] =xS*(Sx*Sz-Cx*Sy*Cz); TRS->M[2][1] =yS*(Sx*Cz+Cx*Sy*Sz); TRS->M[2][2] = zS*Cx*Cy;	TRS->M[2][3] =t->z;
  TRS->M[3][0] =(Scalar)0.0;  		 TRS->M[3][1] = (Scalar)0.0;		TRS->M[3][2] =(Scalar)0.0;	TRS->M[3][3] =(Scalar)1.0;
}

